' rs232.S
' *****************************************************
' * Full duplex serial driver                         *
' * Optimized for 7 bit non-zero ASCII data           *
' *****************************************************

' To build the rs232 class library
' propeller-elf-gcc -Os -c rs232.S rs232lib.c
' propeller-elf-ar rs librs232.a rs232lib.o rs232.o

' To invert receiver change iST to if_c, iSP to if_nc, and iRX to muxnc
' To invert transmitter change iTx to muxnc

'------------------------------------------------------------------------------
' Store address of axii.S's hub ram binary in a C variable called, axii_S

#ifdef __GNUC__  
#define WA 4
' note: gcc/C requires mangled names with extra _ prefixed for linker
            .global __load_start_rs232_cog 
            .global _rs232_S 
_rs232_S	long    __load_start_rs232_cog
            .section .rs232.cog , "ax"
            .compress off
            .pasm       ' doesn't work in simple ide 1.0
'          fcache #( _end - _start ) ' doesn't work in simple ide 1.0
#else
#define WA 1
DAT ' code segment
' Catalina Export rs232_S  <- this comment is acutally a binder directive
C_rs232_S_ ' place holder in hub ram.
#endif

#define ASSEMBLER
#include "rs232.h"
#define nop2 if_never jmp #0

' BSS variables #defined so they don't take up linker space...

#define pTx			( _end+WA*0   )
#define pRx			( _end+WA*1   )
#define pTxe		( _end+WA*2   )
#define pRxe		( _end+WA*3   )
#define bitQtime    ( _end+WA*4   )
#define bitQQtime   ( _end+WA*5   )
#define txPinMask   ( _end+WA*6   )
#define rxPinMask   ( _end+WA*7   )
#define mode        ( _end+WA*8   )

#define txState		( _end+WA*9   )
#define rxByte		( _end+WA*10  )
#define txNext		( _end+WA*11  )
#define txByte		( _end+WA*12  )
#define time		( _end+WA*13  )
#define temp		( _end+WA*14  )
#define zero        ( _end+WA*15  )
_start
			.org	0

bssClear	mov		$1EF,#0			' clear top of memory down
			sub		bssClear,oneDest	' decrement index
			tjnz	_end,#bssClear	' loop until cleared high memory

			mov		temp, PAR		' reverse copy initialization data
init		rdlong	mode,temp
			sub		init,oneDest	' Servo rxPinMask location
erase		wrlong	zero,temp		' erase as we go.
			sub		temp,#4
			tjz		pTxe,#init
			mov		erase,#0		' stop erasing
			tjz		pTx,#init

			mov		OUTA, txPinMask ' Make transmit pin in idle state
			mov		DIRA, txPinMask ' Only turn on transmit pin

			mov		time, cnt
			add		time, bitQtime

' -------------------- Main loop --------------------------------------------
' Reception can be aligned to any one of three offset time slots
' Generally, 1+1/3 bit time from zero crossing is set as the sample point
' Transmission always happens at the same time, regardless.

TS1			waitcnt	time, bitQtime
			jmpret	txState, rxState		' do receive
			rdbyte	txNext,pTx wz		' Read transmit hub data
	if_nz	or		txNext,#$80			' Extend data with stop bit

TS2			waitcnt	time,bitQtime 
			jmpret	txState, rxState	' do receive
			test	txNext, #$80 wc		' if New data, C=1
			cmp		txByte, #0 wz       ' if buffer empty z=1
 if_c_and_z	wrbyte	zero, pTx			' clear used transmit hub data

TS3			waitcnt	time,bitQQtime 
			jmpret	txState, rxState		' do receive

' /\/\/\/\/\/\/\/\/\/\/\/\  TRANSMITTER  /\/\/\/\/\/\/\/\/\/\/\/\/\/ '
			tjz	txByte,#NextTx			' If data to transmit, do it.

			'' Terminate after shifting out the stop bit

Tx			shr		txByte,#1 wz,wc		' get a bit, highest bit ignored
iTx			muxc	OUTA, txPinMask		' send bit, if anything to send
 if_c_and_z	add		pTx,#1				' Goto next byte on send completed
			jmp		#TS1				

			'' Get next transmit byte into transmit buffer.
NextTx		mov		txByte,txNext wz	' Transfer new byte
			muxz	OUTA,txPinMask		' Transmit start bit
			cmp		pTx,pTxe wz			' roll Tx buffer when appropriate 
	if_z	sub		pTx,#SZ_TX			' ...
			jmp		#TS1

' /\/\/\/\/\/\/\/\/\/\/\/\  RECEIVER  /\/\/\/\/\/\/\/\/\/\/\/\/\/\/ '
' -------------------- Receve stop and start code ---------------------------

recv		wrbyte	rxByte, pRx 		' save received data
			jmpret	rxState, txState	
	
WaitStart	test	rxPinMask,INA wc	' Find start bit=0
iST	if_nc	mov		rxState, #Delay0	' If start, receive @ 1+1/3 bit time
			jmp		txState				' Delay 1/3 bit time, try again

Delay0		mov		rxByte,rxByte wz	' See if old data saved requries move.
	if_nz	add		pRx,#1
			jmpret	rxState,txState

Delay1		mov		rxByte, #$40		' Prepare for new 7 bit data input.
			nop
Delay1B		jmpret	rxState, txState

Delay2		cmp		pRx,pRxe wz			' Insure rollover of buffer controlled
	if_z	sub		pRx,#SZ_RX
			jmpret  rxState, txState

sample		test	rxPinMask,INA wz    ' Shift in the data
iRX	 if_nz	or		rxByte, #$80		' Shift bit in at 8th position
			jmpret	rxState, txState    
			
Delay1A		shr		rxByte, #1 wc		' When start bit shifts out,recv finish
 if_nc		jmp		#Delay1B			' Otherwise do Delay2 again.
			jmpret	rxState, txState	' 

WaitStop	test	rxPinMask,INA wz	' Look for stop bit, 1 cycle early.
	if_nz	mov		rxState,#recv		' When found can save data
			jmp		txState				' Continue transmitting

'' -------------------- Initialized data constants ----------------------

rxState		long	WaitStart
oneDest		long	$200
_end
